import json
import logging
import os
from datetime import datetime
from typing import Dict

import numpy as np
import pandas as pd
import pytest
from sklearn.metrics import confusion_matrix, precision_score, recall_score

from datahub_classify.helper_classes import ColumnInfo, Metadata
from datahub_classify.infotype_predictor import predict_infotypes
from datahub_classify.reference_input import input1 as input_dict

logger = logging.getLogger(__name__)

current_wdr = os.path.dirname(os.path.abspath(__file__))
input_data_dir = os.path.join(current_wdr, "datasets")
input_jsons_dir = os.path.join(current_wdr, "expected_output")
confidence_threshold = 0.6
update_confidence_slabs_json = False

logging_directory = os.path.join(current_wdr, "logs", "logs.log")

infotypes_to_use = [
    "Street_Address",
    "Gender",
    "Credit_Debit_Card_Number",
    "Email_Address",
    "Phone_Number",
    "Full_Name",
    "Age",
    "IBAN",
    "Vehicle_Identification_Number",
    "US_Social_Security_Number",
    "IP_Address_v4",
    "IP_Address_v6",
    "Swift_Code",
    "US_Driving_License_Number",
]


def get_public_data(input_data_path):
    logger.info("===============%s=================" % input_data_path)
    dataset_dict = {}
    for root, dirs, files in os.walk(input_data_path):
        for i, filename in enumerate(files):
            if filename.endswith(".csv"):
                dataset_name = filename.replace(".csv", "")
                dataset_dict[dataset_name] = pd.read_csv(os.path.join(root, filename))
            elif filename.endswith(".xlsx"):
                dataset_name = filename.replace(".xlsx", "")
                dataset_dict[dataset_name] = pd.read_excel(os.path.join(root, filename))
    return dataset_dict


def populate_column_info_list(public_data_list):
    column_info_list = []
    actual_labels = []
    for i, (dataset_name, data) in enumerate(public_data_list.items()):
        for col in data.columns:
            fields = {
                "Name": col,
                "Description": f"This column contains name of the {col}",
                "Datatype": "str",
                "Dataset_Name": dataset_name,
            }
            metadata = Metadata(fields)
            if len(data[col].dropna()) > 1000:
                values = data[col].dropna().values[:1000]
            else:
                values = data[col].dropna().values
            col_info = ColumnInfo(metadata, values)
            column_info_list.append(col_info)
            actual_labels.append(col)
    return column_info_list


def get_public_data_expected_output(public_data_list, infotypes_to_use):
    with open(os.path.join(input_jsons_dir, "expected_infotypes_IDEAL.json")) as file:
        expected_output_ideal = json.load(file)
    with open(
        os.path.join(input_jsons_dir, "expected_infotypes_UNIT_TESTING.json")
    ) as file:
        expected_output_unit_testing = json.load(file)
    with open(
        os.path.join(input_jsons_dir, "expected_infotypes_confidence_slabs.json")
    ) as file:
        expected_infotypes_confidence_slabs = json.load(file)

    for dataset in public_data_list.keys():
        for col in public_data_list[dataset].columns:
            if (col not in expected_output_ideal[dataset].keys()) or (
                expected_output_ideal[dataset][col] not in infotypes_to_use
            ):
                expected_output_ideal[dataset][col] = "no_infotype"

            if not expected_infotypes_confidence_slabs.get(dataset, None):
                expected_infotypes_confidence_slabs[dataset] = dict()
            if (col not in expected_infotypes_confidence_slabs[dataset].keys()) or (
                expected_output_ideal[dataset][col] not in infotypes_to_use
            ):
                expected_infotypes_confidence_slabs[dataset][col] = 0.0

            if col not in expected_output_unit_testing[dataset].keys():
                expected_output_unit_testing[dataset][col] = "no_infotype"

    return (
        expected_output_ideal,
        expected_output_unit_testing,
        expected_infotypes_confidence_slabs,
    )


def get_best_infotype_pred(
    public_data_list,
    confidence_threshold,
    expected_output_unit_testing,
    update_confidence_slabs_json=False,
):
    with open(
        os.path.join(input_jsons_dir, "expected_infotypes_confidence_slabs.json")
    ) as filename:
        old_confidence_slabs = json.load(filename)
    column_info_list = populate_column_info_list(public_data_list)
    column_info_pred_list = predict_infotypes(
        column_info_list, confidence_threshold, input_dict, infotypes_to_use
    )
    public_data_predicted_infotype: Dict[str, dict] = dict()
    # get_thresholds_for_unit_test = dict()
    public_data_predicted_infotype_confidence: Dict[str, dict] = dict()
    for dataset in public_data_list.keys():
        public_data_predicted_infotype[dataset] = dict()
        if not old_confidence_slabs.get(dataset):
            old_confidence_slabs[dataset] = dict()
        public_data_predicted_infotype_confidence[dataset] = dict()
        for col in public_data_list[dataset].columns:
            for col_info in column_info_pred_list:
                if (
                    col_info.metadata.name == col
                    and col_info.metadata.dataset_name == dataset
                ):
                    public_data_predicted_infotype[dataset][col] = "no_infotype"
                    if (
                        col_info.infotype_proposals
                        and len(col_info.infotype_proposals) > 0
                    ):
                        highest_confidence_level: float = 0
                        infotype_assigned = None
                        for i in range(len(col_info.infotype_proposals)):
                            if (
                                col_info.infotype_proposals[i].confidence_level
                                > highest_confidence_level
                            ):
                                # get_thresholds_for_unit_test[dataset][col] = col_info.infotype_proposals[i].confidence_level
                                highest_confidence_level = col_info.infotype_proposals[
                                    i
                                ].confidence_level
                                infotype_assigned = col_info.infotype_proposals[
                                    i
                                ].infotype
                        if not old_confidence_slabs[dataset].get(col):
                            old_confidence_slabs[dataset][col] = (
                                np.floor(highest_confidence_level * 10) / 10
                            )
                        public_data_predicted_infotype[dataset][col] = infotype_assigned
                        public_data_predicted_infotype_confidence[dataset][
                            col
                        ] = highest_confidence_level

                        # TODO: what is the use of following condition?
                        if expected_output_unit_testing[dataset][col] not in (
                            infotypes_to_use + ["no_infotype"]
                        ):
                            expected_output_unit_testing[dataset][
                                col
                            ] = infotype_assigned
                    else:
                        if expected_output_unit_testing[dataset][col] not in (
                            infotypes_to_use + ["no_infotype"]
                        ):
                            expected_output_unit_testing[dataset][col] = "no_infotype"
                        public_data_predicted_infotype_confidence[dataset][col] = 0.0
    if update_confidence_slabs_json:
        with open(
            os.path.join(input_jsons_dir, "expected_infotypes_confidence_slabs.json"),
            "w",
        ) as filename:
            json.dump(old_confidence_slabs, filename, indent=4)
    return (
        public_data_predicted_infotype,
        expected_output_unit_testing,
        public_data_predicted_infotype_confidence,
    )


def get_pred_exp_infotype_mapping(
    public_data_predicted_infotype,
    public_data_expected_infotype,
    expected_infotypes_confidence_slabs,
    public_data_predicted_infotype_confidence,
):
    mapping = []
    for dataset in public_data_predicted_infotype.keys():
        for col in public_data_predicted_infotype[dataset].keys():
            mapping.append(
                (
                    dataset,
                    col,
                    public_data_predicted_infotype[dataset][col],
                    public_data_expected_infotype[dataset][col],
                    public_data_predicted_infotype_confidence[dataset][col],
                    expected_infotypes_confidence_slabs[dataset][col],
                )
            )
    return mapping


def get_prediction_statistics(mapping, infotypes_to_use, confidence_threshold):
    infotypes_to_use.append("no_infotype")
    all_infotypes = infotypes_to_use
    y_true = [s[3] for s in mapping]
    y_pred = [s[2] for s in mapping]
    prediction_stats = pd.DataFrame()
    prediction_stats["Infotype"] = all_infotypes
    prediction_stats["Precision"] = np.round(
        precision_score(y_true, y_pred, average=None, labels=all_infotypes), 2
    )
    prediction_stats["Recall"] = np.round(
        recall_score(y_true, y_pred, average=None, labels=all_infotypes), 2
    )
    df_confusion_matrix = pd.DataFrame(
        confusion_matrix(y_true, y_pred, labels=all_infotypes),
        columns=[info + "_predicted" for info in all_infotypes],
        index=[info + "_actual" for info in all_infotypes],
    )
    logger.info("*************Prediction Statistics***************************")
    logger.info(prediction_stats)
    logger.info("********************")
    logger.info(df_confusion_matrix)
    prediction_stats.to_csv(
        f"Prediction_statistics_{confidence_threshold}.csv", index=False
    )
    df_confusion_matrix.to_csv(f"confusion_matrix_{confidence_threshold}.csv")


# TODO: think of adding 'if __name__ == “main”' block for following executable code
# if __name__ == '__main__':
logger.info("+++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
logger.info("--------------------STARTING RUN--------------------  ")
logger.info("+++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
logger.info(
    f"Start Time --->  {datetime.now()}",
)

public_data_list = get_public_data(input_data_dir)
(
    expected_output_ideal,
    expected_output_unit_testing,
    expected_infotypes_confidence_slabs,
) = get_public_data_expected_output(public_data_list, infotypes_to_use)
(
    public_data_predicted_infotype,
    expected_output_unit_testing,
    public_data_predicted_infotype_confidence,
) = get_best_infotype_pred(
    public_data_list,
    confidence_threshold,
    expected_output_unit_testing,
    update_confidence_slabs_json,
)
infotype_mapping_ideal = get_pred_exp_infotype_mapping(
    public_data_predicted_infotype,
    expected_output_ideal,
    expected_infotypes_confidence_slabs,
    public_data_predicted_infotype_confidence,
)
infotype_mapping_unit_testing = get_pred_exp_infotype_mapping(
    public_data_predicted_infotype,
    expected_output_unit_testing,
    expected_infotypes_confidence_slabs,
    public_data_predicted_infotype_confidence,
)
get_prediction_statistics(
    infotype_mapping_ideal, infotypes_to_use, confidence_threshold
)


@pytest.mark.parametrize(
    "dataset_name,column_name, predicted_output, expected_output,"
    "predicted_output_confidence, expected_confidence_slab",
    [(a, b, c, d, e, f) for a, b, c, d, e, f in infotype_mapping_unit_testing],
)
def test_public_datasets(
    dataset_name,
    column_name,
    predicted_output,
    expected_output,
    predicted_output_confidence,
    expected_confidence_slab,
):
    assert (
        predicted_output == expected_output
    ), f"Test1 failed for column '{column_name}' in {dataset_name}"
    assert predicted_output_confidence >= expected_confidence_slab, (
        f"Test2 failed for column '{column_name}' in " f"{dataset_name}"
    )



import ipaddress
import logging
import re
import string
from typing import Any, Dict, List, Tuple, cast

import numpy as np
import phonenumbers
import spacy
from schwifty import BIC, IBAN
from stdnum.us import ssn as us_ssn
from vininfo import Vin

from datahub_classify.constants import (
    DATATYPE,
    DESCRIPTION,
    NAME,
    PREDICTION_FACTORS_AND_WEIGHTS,
    PREDICTION_TYPE,
    REGEX,
    TYPE,
    VALUES,
)
from datahub_classify.helper_classes import DebugInfo, Metadata
from datahub_classify.infotype_utils import (
    detect_named_entity_spacy,
    match_datatype,
    match_regex,
    match_regex_for_values,
)

logger = logging.getLogger(__name__)
spacy_model_name = "en_core_web_sm"
try:
    nlp_english = spacy.load(spacy_model_name)
except OSError:
    logger.info("Downloading language model for the spaCy")
    from spacy.cli import download

    download(spacy_model_name)
    nlp_english = spacy.load(spacy_model_name)
spacy_models_list = [nlp_english]


def compute_name_description_dtype_score(
    metadata: Metadata, config: Dict[str, Dict], debug_info: DebugInfo
) -> DebugInfo:
    prediction_factors_weights = config[PREDICTION_FACTORS_AND_WEIGHTS]
    # Name Logic
    if prediction_factors_weights.get(NAME, 0) > 0:
        if not metadata.name or not metadata.name.strip():
            # TODO: Add message "Name is None/Blank" in error flag
            # debug_info.name = f"0.0 (Blank {NAME} Metadata)"
            pass
        else:
            debug_info.name = match_regex(metadata.name, config[NAME][REGEX])

    # Description_Logic
    if prediction_factors_weights.get(DESCRIPTION, 0) > 0:
        if not metadata.description or not metadata.description.strip():
            # TODO: Add message "Description is None/Blank" in error flag
            # debug_info.description = f"0.0 (Blank {DESCRIPTION} Metadata)"
            pass
        else:
            debug_info.description = match_regex(
                metadata.description, config[DESCRIPTION][REGEX]
            )

    # Datatype_Logic
    if prediction_factors_weights.get(DATATYPE, 0) > 0:
        if not metadata.datatype or not metadata.datatype.strip():
            # TODO: Add message "Datatype is None/Blank" in error flag
            # debug_info.datatype = f"0.0 (Blank {DATATYPE} Metadata)"
            pass
        else:
            debug_info.datatype = match_datatype(
                metadata.datatype, config[DATATYPE][TYPE]
            )
    return debug_info


def compute_overall_confidence(debug_info: DebugInfo, config: Dict[str, Dict]) -> float:
    prediction_factors_weights = config[PREDICTION_FACTORS_AND_WEIGHTS]
    prediction_factors_weights = {
        k.lower(): v for k, v in prediction_factors_weights.items()
    }
    confidence_level = 0
    for key, value in vars(debug_info).items():
        if value and type(value) != str:
            confidence_level += prediction_factors_weights[key] * value
    confidence_level = np.round(confidence_level, 2)
    return confidence_level


def inspect_for_email_address(
    metadata: Metadata, values: List[Any], config: Dict[str, Dict]
) -> Tuple[float, DebugInfo]:
    prediction_factors_weights = config[PREDICTION_FACTORS_AND_WEIGHTS]
    debug_info = DebugInfo()
    # Value Logic
    if prediction_factors_weights.get(VALUES, 0) > 0:
        values_score = 0.0
        try:
            if config[VALUES][PREDICTION_TYPE] == "regex":
                values_score = match_regex_for_values(values, config[VALUES][REGEX])
            elif config[VALUES][PREDICTION_TYPE] == "library":
                raise Exception(
                    "Currently prediction type 'library' is not supported for infotype Email Address"
                )
            else:
                raise Exception(
                    f"Inappropriate Prediction type {config[VALUES][PREDICTION_TYPE]}"
                )
        except Exception as e:
            logger.error(f"Column {metadata.name} failed due to {e}")
        values_score = np.round(values_score, 2)
        debug_info.values = values_score

    debug_info = compute_name_description_dtype_score(metadata, config, debug_info)
    confidence_level = compute_overall_confidence(debug_info, config)
    return confidence_level, debug_info


def inspect_for_street_address(
    metadata: Metadata, values: List[Any], config: Dict[str, Dict]
) -> Tuple[float, DebugInfo]:  # noqa: C901
    prediction_factors_weights = config[PREDICTION_FACTORS_AND_WEIGHTS]
    debug_info = DebugInfo()

    # Values logic
    if prediction_factors_weights.get(VALUES, 0) > 0:
        values_score = 0.0
        try:
            if config[VALUES][PREDICTION_TYPE] == "regex":
                values_score = match_regex_for_values(values, config[VALUES][REGEX])
            elif config[VALUES][PREDICTION_TYPE] == "library":
                entity_count: float = 0
                entities_of_interest = ["FAC", "LOC", "ORG"]
                weight = 1.5
                for value in values:
                    try:
                        if detect_named_entity_spacy(
                            spacy_models_list, entities_of_interest, value
                        ):
                            entity_count += weight
                    except Exception:
                        pass
                entities_score = entity_count / len(values)
                values_score = np.minimum(entities_score, 1)
            else:
                raise Exception(
                    f"Inappropriate values_prediction_type {config[VALUES][PREDICTION_TYPE]}"
                )
        except Exception as e:
            logger.error(f"Column {metadata.name} failed due to {e}")
            pass
        values_score = np.round(values_score, 2)
        debug_info.values = values_score

    debug_info = compute_name_description_dtype_score(metadata, config, debug_info)
    confidence_level = compute_overall_confidence(debug_info, config)
    return confidence_level, debug_info


def inspect_for_gender(
    metadata: Metadata, values: List[Any], config: Dict[str, Dict]
) -> Tuple[float, DebugInfo]:  # noqa: C901
    prediction_factors_weights = config[PREDICTION_FACTORS_AND_WEIGHTS]
    debug_info = DebugInfo()

    # Value Logic
    if prediction_factors_weights.get(VALUES, 0) > 0:
        values_score = 0.0
        try:
            if config[VALUES][PREDICTION_TYPE] == "regex":
                values_score = match_regex_for_values(values, config[VALUES][REGEX])
            elif config[VALUES][PREDICTION_TYPE] == "library":
                raise Exception(
                    "Currently prediction type 'library' is not supported for infotype Gender"
                )
            else:
                raise Exception(
                    f"Inappropriate values_prediction_type {config[VALUES][PREDICTION_TYPE]}"
                )
        except Exception as e:
            logger.error(f"Column {metadata.name} failed due to {e}")
        values_score = np.round(values_score, 2)
        debug_info.values = values_score

    debug_info = compute_name_description_dtype_score(metadata, config, debug_info)

    try:
        if (
            prediction_factors_weights.get(NAME, 0) > 0
            and debug_info.name is not None
            and abs(1 - debug_info.name) < 1e-10
            and prediction_factors_weights.get(VALUES, 0) > 0
            and debug_info.values == 0.0
        ):
            num_unique_values = len(np.unique(values))
            if num_unique_values < 5:
                debug_info.values = 0.9
    except Exception as e:
        logger.error(
            f"Failed to evaluate unique values when name_score==1 for Gender due to {e}"
        )

    confidence_level = compute_overall_confidence(debug_info, config)
    return confidence_level, debug_info


def inspect_for_credit_debit_card_number(
    metadata: Metadata, values: List[Any], config: Dict[str, Dict]
) -> Tuple[float, DebugInfo]:
    prediction_factors_weights = config[PREDICTION_FACTORS_AND_WEIGHTS]
    debug_info = DebugInfo()

    # Value Logic
    if prediction_factors_weights.get(VALUES, 0) > 0:
        values_score = 0.0
        try:
            values = [str(s).lower() for s in values]
            values_cleaned = []
            for value in values:
                string_cleaned = re.sub(r"[ _-]+", "", value)
                values_cleaned.append(string_cleaned)
            if config[VALUES][PREDICTION_TYPE] == "regex":
                values_score = match_regex_for_values(
                    values_cleaned, config[VALUES][REGEX]
                )
            elif config[VALUES][PREDICTION_TYPE] == "library":
                raise Exception(
                    "Currently prediction type 'library' is not supported for infotype Credit Card Number"
                )
            else:
                raise Exception(
                    f"Inappropriate values_prediction_type {config[VALUES][PREDICTION_TYPE]}"
                )
        except Exception as e:
            logger.error(f"Column {metadata.name} failed due to {e}")
        values_score = np.round(values_score, 2)
        debug_info.values = values_score

    debug_info = compute_name_description_dtype_score(metadata, config, debug_info)
    confidence_level = compute_overall_confidence(debug_info, config)
    return confidence_level, debug_info


def inspect_for_phone_number(
    metadata: Metadata, values: List[Any], config: Dict[str, Dict]
) -> Tuple[float, DebugInfo]:  # noqa: C901
    prediction_factors_weights = config[PREDICTION_FACTORS_AND_WEIGHTS]
    debug_info = DebugInfo()

    # fmt: off
    # TODO: shall we have these country codes in config?
    iso_codes = ["AF", "AX", "AL", "DZ", "AS", "AD", "AO", "AI", "AQ", "AG", "AR",
                 "AM", "AW", "AU", "AT", "AZ", "BS", "BH", "BD", "BB", "BY", "BE",
                 "BZ", "BJ", "BM", "BT", "BO", "BQ", "BA", "BW", "BV", "BR", "IO",
                 "BN", "BG", "BF", "BI", "CV", "KH", "CM", "CA", "KY", "CF", "TD",
                 "CL", "CN", "CX", "CC", "CO", "KM", "CG", "CD", "CK", "CR", "CI",
                 "HR", "CU", "CW", "CY", "CZ", "DK", "DJ", "DM", "DO", "EC", "EG",
                 "SV", "GQ", "ER", "EE", "ET", "FK", "FO", "FJ", "FI", "FR", "GF",
                 "PF", "TF", "GA", "GM", "GE", "DE", "GH", "GI", "GR", "GL", "GD",
                 "GP", "GU", "GT", "GG", "GN", "GW", "GY", "HT", "HM", "VA", "HN",
                 "HK", "HU", "IS", "IN", "ID", "IR", "IQ", "IE", "IM", "IL", "IT",
                 "JM", "JP", "JE", "JO", "KZ", "KE", "KI", "KP", "KR", "KW", "KG",
                 "LA", "LV", "LB", "LS", "LR", "LY", "LI", "LT", "LU", "MO", "MK",
                 "MG", "MW", "MY", "MV", "ML", "MT", "MH", "MQ", "MR", "MU", "YT",
                 "MX", "FM", "MD", "MC", "MN", "ME", "MS", "MA", "MZ", "MM", "NA",
                 "NR", "NP", "NL", "NC", "NZ", "NI", "NE", "NG", "NU", "NF", "MP",
                 "NO", "OM", "PK", "PW", "PS", "PA", "PG", "PY", "PE", "PH", "PN",
                 "PL", "PT", "PR", "QA", "RE", "RO", "RU", "RW", "BL", "SH", "KN",
                 "LC", "MF", "PM", "VC", "WS", "SM", "ST", "SA", "SN", "RS", "SC",
                 "SL", "SG", "SX", "SK", "SI", "SB", "SO", "ZA", "GS", "SS", "ES",
                 "LK", "SD", "SR", "SJ", "SZ", "SE", "CH", "SY", "TW", "TJ", "TZ",
                 "TH", "TL", "TG", "TK", "TO", "TT", "TN", "TR", "TM", "TC", "TV",
                 "UG", "UA", "AE", "GB", "US", "UM", "UY", "UZ", "VU", "VE", "VN",
                 "VG", "VI", "WF", "EH", "YE", "ZM", "ZW"]
    # fmt: on
    # Values logic
    if prediction_factors_weights.get(VALUES, 0) > 0:
        values_score: float = 0.0
        try:
            if config[VALUES][PREDICTION_TYPE] == "regex":
                values_score = match_regex_for_values(values, config[VALUES][REGEX])
            elif config[VALUES][PREDICTION_TYPE] == "library":
                valid_phone_numbers_count = 0
                for value in values:
                    try:
                        for code in iso_codes:
                            parsed_number = phonenumbers.parse(value, code)
                            if phonenumbers.is_possible_number(parsed_number):
                                valid_phone_numbers_count += 1
                                break
                    except Exception:
                        pass
                values_score = valid_phone_numbers_count / len(values)
            else:
                raise Exception(
                    f"Inappropriate values_prediction_type {config[VALUES][PREDICTION_TYPE]}"
                )

        except Exception as e:
            logger.error(f"Column {metadata.name} failed due to {e}")

        values_score = np.round(values_score, 2)
        debug_info.values = values_score

    debug_info = compute_name_description_dtype_score(metadata, config, debug_info)
    confidence_level = compute_overall_confidence(debug_info, config)
    return confidence_level, debug_info


def inspect_for_full_name(
    metadata: Metadata, values: List[Any], config: Dict[str, Dict]
) -> Tuple[float, DebugInfo]:  # noqa: C901
    prediction_factors_weights = config[PREDICTION_FACTORS_AND_WEIGHTS]
    debug_info = DebugInfo()

    # Values logic
    if prediction_factors_weights.get(VALUES, 0) > 0:
        values_score = 0.0
        try:
            if config[VALUES][PREDICTION_TYPE] == "regex":
                values_score = match_regex_for_values(values, config[VALUES][REGEX])
            elif config[VALUES][PREDICTION_TYPE] == "library":
                entity_count = 0
                entities_of_interest = ["PERSON"]
                weight = 1
                for value in values:
                    try:
                        if (
                            len(value) <= 50
                            and len(re.split(rf"[\s{string.punctuation}]+", value)) >= 2
                        ):
                            if detect_named_entity_spacy(
                                spacy_models_list, entities_of_interest, value
                            ):
                                entity_count += weight
                    except Exception:
                        pass
                entities_score = entity_count / len(values)
                values_score = np.minimum(entities_score, 1)
            else:
                raise Exception(
                    f"Inappropriate values_prediction_type {config[VALUES][PREDICTION_TYPE]}"
                )
        except Exception as e:
            logger.error(f"Column {metadata.name} failed due to {e}")
        values_score = np.round(values_score, 2)
        debug_info.values = values_score

    debug_info = compute_name_description_dtype_score(metadata, config, debug_info)

    try:
        if (
            prediction_factors_weights.get(NAME, 0) > 0
            and debug_info.name is not None
            and abs(1 - debug_info.name) < 1e-10
            and prediction_factors_weights.get(VALUES, 0) > 0
            and 0.5 > cast(float, debug_info.values) > 0.1
        ):
            debug_info.values = 0.8
    except Exception as e:
        logger.error(f"Column {metadata.name} failed due to {e}")

    confidence_level = compute_overall_confidence(debug_info, config)
    return confidence_level, debug_info


def inspect_for_age(
    metadata: Metadata, values: List[Any], config: Dict[str, Dict]
) -> Tuple[float, DebugInfo]:  # noqa: C901
    prediction_factors_weights = config[PREDICTION_FACTORS_AND_WEIGHTS]
    debug_info = DebugInfo()

    # Values logic
    if prediction_factors_weights.get(VALUES, 0) > 0:
        values_score: float = 0.0
        try:
            if config[VALUES][PREDICTION_TYPE] == "regex":
                values_score = match_regex_for_values(values, config[VALUES][REGEX])
            elif config[VALUES][PREDICTION_TYPE] == "library":
                try:
                    # Check if column is convertible to int dtype
                    int_col = [int(s) for s in values]
                    max_val = np.percentile(int_col, 95)
                    min_val = np.percentile(int_col, 5)
                    num_unique = len(np.unique(int_col))
                    if max_val <= 120 and min_val > 0:
                        values_score += 0.4
                        if num_unique >= np.minimum(len(values) / 10, 40):
                            values_score += 0.3
                    else:
                        values_score = 0
                except Exception:
                    pass
            else:
                raise Exception(
                    f"Inappropriate values_prediction_type {config[VALUES][PREDICTION_TYPE]}"
                )
        except Exception as e:
            logger.error(f"Column {metadata.name} failed due to {e}")
        debug_info.values = values_score

    debug_info = compute_name_description_dtype_score(metadata, config, debug_info)
    confidence_level = compute_overall_confidence(debug_info, config)
    return confidence_level, debug_info


def inspect_for_iban(
    metadata: Metadata, values: List[Any], config: Dict[str, Dict]
) -> Tuple[float, DebugInfo]:  # noqa: C901
    prediction_factors_weights = config[PREDICTION_FACTORS_AND_WEIGHTS]
    debug_info = DebugInfo()

    # Value Logic
    if prediction_factors_weights.get(VALUES, 0) > 0:
        values_score = 0.0
        try:
            if config[VALUES][PREDICTION_TYPE] == "regex":
                values_score = match_regex_for_values(values, config[VALUES][REGEX])
            elif config[VALUES][PREDICTION_TYPE] == "library":
                iban_score = 0.0
                for val in values:
                    try:
                        if IBAN(val, allow_invalid=True).is_valid:
                            iban_score += 1.0
                    except Exception:
                        pass
                values_score = iban_score / len(values)
            else:
                raise Exception(
                    f"Inappropriate values_prediction_type {config[VALUES][PREDICTION_TYPE]}"
                )
        except Exception as e:
            logger.error(f"Column {metadata.name} failed due to {e}")
        values_score = np.round(values_score, 2)
        debug_info.values = values_score

    debug_info = compute_name_description_dtype_score(metadata, config, debug_info)
    confidence_level = compute_overall_confidence(debug_info, config)
    return confidence_level, debug_info


def inspect_for_vehicle_identification_number(
    metadata: Metadata, values: List[Any], config: Dict[str, Dict]
) -> Tuple[float, DebugInfo]:  # noqa: C901
    prediction_factors_weights = config[PREDICTION_FACTORS_AND_WEIGHTS]
    debug_info = DebugInfo()

    # Value Logic
    if prediction_factors_weights.get(VALUES, 0) > 0:
        values_score = 0.0
        try:
            if config[VALUES][PREDICTION_TYPE] == "regex":
                values_score = match_regex_for_values(values, config[VALUES][REGEX])
            elif config[VALUES][PREDICTION_TYPE] == "library":
                vin_score = 0
                for val in values:
                    try:
                        # Vin constructor implicitly validates the VIN
                        _ = Vin(val)
                        vin_score += 1
                    except Exception:
                        pass
                values_score = vin_score / len(values)
            else:
                raise Exception(
                    f"Inappropriate values_prediction_type {config[VALUES][PREDICTION_TYPE]}"
                )
        except Exception as e:
            logger.error(f"Column {metadata.name} failed due to {e}")
        values_score = np.round(values_score, 2)
        debug_info.values = values_score

    debug_info = compute_name_description_dtype_score(metadata, config, debug_info)
    confidence_level = compute_overall_confidence(debug_info, config)
    return confidence_level, debug_info


def inspect_for_ip_address_v4(
    metadata: Metadata, values: List[Any], config: Dict[str, Dict]
) -> Tuple[float, DebugInfo]:  # noqa: C901
    prediction_factors_weights = config[PREDICTION_FACTORS_AND_WEIGHTS]
    debug_info = DebugInfo()

    # Value Logic
    if prediction_factors_weights.get(VALUES, 0) > 0:
        values_score = 0.0
        try:
            if config[VALUES][PREDICTION_TYPE] == "regex":
                values_score = match_regex_for_values(values, config[VALUES][REGEX])
            elif config[VALUES][PREDICTION_TYPE] == "library":
                count = 0
                for value in values:
                    try:
                        _ = ipaddress.IPv4Address(str(value))
                        count = count + 1
                    except Exception:
                        pass
                values_score = count / len(values)
            else:
                raise Exception(
                    f"Inappropriate values_prediction_type {config[VALUES][PREDICTION_TYPE]}"
                )
        except Exception as e:
            logger.error(f"Column {metadata.name} failed due to {e}")
        values_score = np.round(values_score, 2)
        debug_info.values = values_score

    debug_info = compute_name_description_dtype_score(metadata, config, debug_info)
    confidence_level = compute_overall_confidence(debug_info, config)
    return confidence_level, debug_info


def inspect_for_ip_address_v6(
    metadata: Metadata, values: List[Any], config: Dict[str, Dict]
) -> Tuple[float, DebugInfo]:  # noqa: C901
    prediction_factors_weights = config[PREDICTION_FACTORS_AND_WEIGHTS]
    debug_info = DebugInfo()

    # Value Logic
    if prediction_factors_weights.get(VALUES, 0) > 0:
        values_score = 0.0
        try:
            if config[VALUES][PREDICTION_TYPE] == "regex":
                values_score = match_regex_for_values(values, config[VALUES][REGEX])
            elif config[VALUES][PREDICTION_TYPE] == "library":
                count = 0
                for value in values:
                    try:
                        _ = ipaddress.IPv6Address(str(value))
                        count = count + 1
                    except Exception:
                        pass
                values_score = count / len(values)
            else:
                raise Exception(
                    f"Inappropriate values_prediction_type {config[VALUES][PREDICTION_TYPE]}"
                )
        except Exception as e:
            logger.error(f"Column {metadata.name} failed due to {e}")
        values_score = np.round(values_score, 2)
        debug_info.values = values_score

    debug_info = compute_name_description_dtype_score(metadata, config, debug_info)
    confidence_level = compute_overall_confidence(debug_info, config)
    return confidence_level, debug_info


def inspect_for_us_driving_license_number(
    metadata: Metadata, values: List[Any], config: Dict[str, Dict]
) -> Tuple[float, DebugInfo]:
    prediction_factors_weights = config[PREDICTION_FACTORS_AND_WEIGHTS]
    debug_info = DebugInfo()

    # Value Logic
    if prediction_factors_weights.get(VALUES, 0) > 0:
        values_score = 0.0
        try:
            if config[VALUES][PREDICTION_TYPE] == "regex":
                values_score = match_regex_for_values(values, config[VALUES][REGEX])
            elif config[VALUES][PREDICTION_TYPE] == "library":
                raise Exception(
                    "Currently prediction type 'library' is not supported for "
                    "infotype US Driving License Number"
                )
            else:
                raise Exception(
                    f"Inappropriate values_prediction_type {config[VALUES][PREDICTION_TYPE]}"
                )
        except Exception as e:
            logger.error(f"Column {metadata.name} failed due to {e}")
        values_score = np.round(values_score, 2)
        debug_info.values = values_score

    debug_info = compute_name_description_dtype_score(metadata, config, debug_info)
    confidence_level = compute_overall_confidence(debug_info, config)
    return confidence_level, debug_info


def inspect_for_us_social_security_number(
    metadata: Metadata, values: List[Any], config: Dict[str, Dict]
) -> Tuple[float, DebugInfo]:  # noqa: C901
    prediction_factors_weights = config[PREDICTION_FACTORS_AND_WEIGHTS]
    debug_info = DebugInfo()

    # Value Logic
    if prediction_factors_weights.get(VALUES, 0) > 0:
        values_score = 0.0
        try:
            if config[VALUES][PREDICTION_TYPE] == "regex":
                values_score = match_regex_for_values(values, config[VALUES][REGEX])
            elif config[VALUES][PREDICTION_TYPE] == "library":
                us_ssn_score = 0
                for val in values:
                    try:
                        if us_ssn.is_valid(val):
                            us_ssn_score += 1
                    except Exception:
                        pass
                values_score = us_ssn_score / len(values)
            else:
                raise Exception(
                    f"Inappropriate values_prediction_type {config[VALUES][PREDICTION_TYPE]}"
                )
        except Exception as e:
            logger.error(f"Column {metadata.name} failed due to {e}")
        values_score = np.round(values_score, 2)
        debug_info.values = values_score

    debug_info = compute_name_description_dtype_score(metadata, config, debug_info)
    confidence_level = compute_overall_confidence(debug_info, config)
    return confidence_level, debug_info


def inspect_for_swift_code(
    metadata: Metadata, values: List[Any], config: Dict[str, Dict]
) -> Tuple[float, DebugInfo]:  # noqa: C901
    prediction_factors_weights = config[PREDICTION_FACTORS_AND_WEIGHTS]
    debug_info = DebugInfo()

    # Value Logic
    if prediction_factors_weights.get(VALUES, 0) > 0:
        values_score = 0.0
        try:
            if config[VALUES][PREDICTION_TYPE] == "regex":
                values_score = match_regex_for_values(values, config[VALUES][REGEX])
            elif config[VALUES][PREDICTION_TYPE] == "library":
                swift_score = 0
                for val in values:
                    try:
                        if BIC(val, allow_invalid=True).is_valid:
                            swift_score += 1
                    except Exception:
                        pass
                values_score = swift_score / len(values)
            else:
                raise Exception(
                    f"Inappropriate values_prediction_type {config[VALUES][PREDICTION_TYPE]}"
                )
        except Exception as e:
            logger.error(f"Column {metadata.name} failed due to {e}")
        values_score = np.round(values_score, 2)
        debug_info.values = values_score

    debug_info = compute_name_description_dtype_score(metadata, config, debug_info)
    confidence_level = compute_overall_confidence(debug_info, config)
    return confidence_level, debug_info



from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional


@dataclass
class InfotypeProposal:
    infotype: str
    confidence_level: float
    debug_info: Dict[str, Any]


@dataclass
class Metadata:
    meta_info: Dict[str, Any]
    name: str = field(init=False)
    description: str = field(init=False)
    datatype: str = field(init=False)
    dataset_name: str = field(init=False)

    def __post_init__(self):
        self.name = self.meta_info.get("Name", None)
        self.description = self.meta_info.get("Description", None)
        self.datatype = self.meta_info.get("Datatype", None)
        self.dataset_name = self.meta_info.get("Dataset_Name", None)


@dataclass
class ColumnInfo:
    metadata: Metadata
    values: List[Any]
    infotype_proposals: Optional[List[InfotypeProposal]] = None


@dataclass
class DebugInfo:
    name: Optional[float] = None
    description: Optional[float] = None
    datatype: Optional[float] = None
    values: Optional[float] = None



import importlib
import logging
from typing import Any, Dict, List, Optional

from datahub_classify.helper_classes import ColumnInfo, InfotypeProposal
from datahub_classify.infotype_utils import perform_basic_checks

logger = logging.getLogger(__name__)


def get_infotype_function_mapping(
    infotypes: Optional[List[str]], global_config: Dict[str, Dict]
) -> Dict[str, Any]:
    from inspect import getmembers, isfunction

    module_name = "datahub_classify.infotype_helper"
    module = importlib.import_module(module_name)
    module_fn_dict = dict(getmembers(module, isfunction))
    infotype_function_map = {}
    if not infotypes:
        infotypes = list(global_config.keys())
    for infotype in infotypes:
        if infotype not in global_config.keys():
            logger.warning(f"Configuration is not available for infotype - {infotype}")
        else:
            fn_name = f"inspect_for_{infotype.lower()}"
            infotype_function_map[infotype] = module_fn_dict[fn_name]
    return infotype_function_map


def predict_infotypes(
    column_infos: List[ColumnInfo],
    confidence_level_threshold: float,
    global_config: Dict[str, Dict],
    infotypes: Optional[List[str]] = None,
) -> List[ColumnInfo]:
    infotype_function_map = get_infotype_function_mapping(infotypes, global_config)
    logger.debug(f"Total columns to be processed --> {len(column_infos)}")
    logger.debug(f"Confidence Level Threshold set to --> {confidence_level_threshold}")
    logger.debug("===========================================================")
    basic_checks_failed_columns = []
    num_cols_with_infotype_assigned = 0
    for column_info in column_infos:
        logger.debug(
            f"processing column: {column_info.metadata.name} -- dataset: {column_info.metadata.dataset_name}"
        )
        # iterate over all infotype functions
        proposal_list = []
        for infotype, infotype_fn in infotype_function_map.items():
            # get the configuration
            config_dict = global_config[infotype]

            # call the infotype prediction function
            column_info.values = [
                val
                for val in column_info.values
                if str(val).strip() not in ["nan", "", "None"]
            ]
            try:
                if perform_basic_checks(
                    column_info.metadata, column_info.values, config_dict, infotype
                ):
                    confidence_level, debug_info = infotype_fn(
                        column_info.metadata, column_info.values, config_dict
                    )
                    if confidence_level > confidence_level_threshold:
                        infotype_proposal = InfotypeProposal(
                            infotype, confidence_level, debug_info
                        )
                        proposal_list.append(infotype_proposal)
                else:
                    basic_checks_failed_columns.append(
                        (column_info.metadata.name, column_info.metadata.dataset_name)
                    )

            except Exception as e:
                # traceback.print_exc()
                logger.warning(f"Failed to extract info type due to {e}")
        if len(proposal_list) > 0:
            num_cols_with_infotype_assigned += 1
        column_info.infotype_proposals = proposal_list
    if len(basic_checks_failed_columns) > 0:
        basic_checks_failed_columns_set = set(basic_checks_failed_columns)
        logger.warning(
            f"Infotype not extracted due to basic checks failure for {len(basic_checks_failed_columns_set)} out of {len(column_infos)} columns.Check DEBUG logs for details"
        )
        logger.debug(
            f"Basic Checks failed for following (column_name, table_name) --> {basic_checks_failed_columns_set}"
        )
    logger.debug(
        f"Infotype assigned to {num_cols_with_infotype_assigned} out of {len(column_infos)} columns"
    )

    return column_infos



import logging
import re
from typing import Any, Dict, List, Optional

from datahub_classify.constants import PREDICTION_FACTORS_AND_WEIGHTS, VALUES
from datahub_classify.helper_classes import Metadata

logger = logging.getLogger(__name__)


# TODO: Exception handling
# Match regex for Name and Description
def match_regex(text_to_match: str, regex_list: List[str]) -> float:
    original_text = text_to_match.lower()
    cleaned_text = "".join(e for e in original_text if e.isalpha())
    match_score: float = 0
    for pattern in regex_list:
        try:
            # TODO: evaluate a case if [A-Za-z] is present in the pattern then it will not give any error,
            # TODO: are there any other cases like above?
            pattern = pattern.lower()
            cleaned_pattern = "".join(e for e in pattern if e.isalpha())
            if (cleaned_pattern == cleaned_text) or (
                re.fullmatch(pattern, original_text)
            ):
                match_score = 1
                break
            # elif re.match(pattern,cleaned_text):  ## revisit later
            #     match_score = 1
            #     break
            elif pattern in original_text:
                match_score = 0.65
            else:
                pass
        except Exception as e:
            logger.error(f"Column Name matching failed due to: {e}")
    return match_score


# Match data type
def match_datatype(dtype_to_match: str, dtype_list: List[str]) -> int:
    dtype_list = [str(s).lower() for s in dtype_list]
    dtype_to_match = dtype_to_match.lower()
    if dtype_to_match in dtype_list:
        match_score = 1
    else:
        match_score = 0
    return match_score


# Match regex for values
def match_regex_for_values(values: List[Any], regex_list: List[str]) -> float:
    values_score_list = []
    length_values = len(values)
    values = [str(x).lower() for x in values]
    for pattern in regex_list:
        try:
            r = re.compile(pattern)
            matches = list(filter(r.fullmatch, values))
            values = [val for val in values if val not in matches]
            values_score_list.append(len(matches))
            if len(values) == 0:
                break
        except Exception as e:
            # TODO: print the exception for debugging purpose
            logger.error(f"Regex match for values failed due to: {e}", exc_info=e)
    values_score = sum(values_score_list) / length_values
    return values_score


def detect_named_entity_spacy(
    spacy_models_list: List, entities_of_interest: List[str], value: str
) -> bool:
    for spacy_model in spacy_models_list:
        doc = spacy_model(value)
        for ent in doc.ents:
            if ent.label_ in entities_of_interest:
                return True
    return False


def perform_basic_checks(
    metadata: Metadata,
    values: List[Any],
    config_dict: Dict[str, Dict],
    infotype: Optional[str] = None,
) -> bool:
    basic_checks_status = True
    minimum_values_threshold = 50
    if (
        config_dict[PREDICTION_FACTORS_AND_WEIGHTS].get(VALUES, None)
        and len(values) < minimum_values_threshold
    ):
        basic_checks_status = False
    # TODO: Add more basic checks
    return basic_checks_status
